#include <stdio.h>
#include <windows.h>
#include "privilege.h"

void print_help() {
    printf("Usage:\n"
           "  wkill <options> [<params>]\n"
           "Options:\n"
           "  -c, --caption <Window Caption>  Caption of the window to be closed\n"
           "  -a, --all                       Kill all matching windows.\n"
           "                                    By default only the 1st found window is closed\n"
           "  -e, --exact                     Matching window captions exactly.\n"
           "                                    By default all captions beginning from <Window Caption>\n"
           "                                    are matched.\n"
           "  -i                              Use case insensetive comparison.\n"
           "  -p, --pid <Process ID>          Kill by process ID (PID).\n"
           "  -n, --process-name <exec name>  Kill by process (executable) name.\n"
           "  -N, --process-name-substr <exec name>\n"
           "                                    Kill by matching part of full process name/path.\n"
           "                                    By default (with -n) module name is matched exactly.\n"
           "  -x, --exit-code <Exit code>     Process exit code (-1 by default).\n"
           "  -l, --list                      Enum (mached) processes without killing.\n"
           "Examples:\n"
           "  wkill -c \"Microsoft Word\"\n"
           "  wkill -p 440\n"
          );
    exit(0);
} // end print_help()

DWORD
EnumProcessPid(
    ULONG pid
    );

BOOLEAN killall = FALSE;
BOOLEAN exact   = FALSE;
BOOLEAN case_sens = TRUE;
BOOLEAN list_only = FALSE;
int kill_count=0;

HMODULE hNtdll = NULL;
HMODULE hPsApi = NULL;


BOOL
CALLBACK
kill_by_caption_EnumWindowsProc(
    HWND hwnd,      // handle to parent window
    LPARAM lParam   // application-defined value
    )
{
//    BOOL    bAttachTh;              // indicates can we attach the process
    DWORD   dwCurThreadId = 0;      // id of current thread
    DWORD   dwCurProcessId = 0;     // id of current process
    //DWORD   dwBadWndThreadId = 0;   // bad window thread id   
    //DWORD   dwBadWndProcId;         // process id of "Security Warning" window

    PWCHAR  WinCap = (PWCHAR)lParam;
    PWCHAR  WinCap2 = NULL;

    ULONG len, len2;

    BOOL res = TRUE;
    BOOLEAN match = FALSE;

    len2 = GetWindowTextLengthW(hwnd);
    if(!len2)
        return TRUE;
    len = wcslen(WinCap);
    WinCap2 = (PWCHAR)GlobalAlloc(GMEM_FIXED, (len2+1)*sizeof(WCHAR));
    if(!WinCap2)
        return TRUE;
    GetWindowTextW(hwnd, WinCap2, len2*sizeof(WCHAR));

    if(len2 < len) {
        len = len2;
    }
    if(!exact) {
        if(case_sens) {
            match = !wcsncmp(WinCap, WinCap2, len);
        } else {
            match = !wcsnicmp(WinCap, WinCap2, len);
        }
    } else {
        if(case_sens) {
            match = !wcscmp(WinCap, WinCap2);
        } else {
            match = !wcsicmp(WinCap, WinCap2);
        }
    }
    if(match) {

        // Get process ID's
        dwCurThreadId  = GetCurrentThreadId();
        dwCurProcessId = GetCurrentProcessId();
        if(list_only) {
            EnumProcessPid(dwCurProcessId);
        } else {
            //dwBadWndThreadId = GetWindowThreadProcessId( hwnd, &dwBadWndProcId);

            //bAttachTh = AttachThreadInput( dwCurThreadId, dwBadWndThreadId, TRUE );

            //SetFocus(hwnd);
        //    if ((SetFocus(hwnd) != NULL) && (bAttachTh))
            {
                //SetActiveWindow( hwnd );
                PostMessage(hwnd, WM_CLOSE, 0, 0);
            }
        }
        if(!killall) {
            res = FALSE;
        }
        kill_count++;
    }
    GlobalFree(WinCap2);

    return res;
} // end kill_by_caption_EnumWindowsProc()

int
kill_by_caption(
    PWCHAR WinCap
    )
{
    EnumWindows(kill_by_caption_EnumWindowsProc, (LONG)WinCap);
    return 0;
} // end kill_by_caption()

int
kill_by_pid(
    ULONG pid,
    ULONG ecode
    )
{
    HANDLE hProcess;

    hProcess=OpenProcess(PROCESS_TERMINATE, FALSE, pid);

    if (hProcess == NULL) {
        printf("Can't open specified process (PID=%d)\n", pid);
        return -1;
    }

    if(!TerminateProcess(hProcess, ecode)) {
        printf("Can't kill process (PID=%d)\n", pid);
        return -1;
    }
    printf("Process (PID=%d) killed\n", pid);
    kill_count++;

    return 0;
} // end kill_by_caption()

ULONG (__stdcall *NtQuerySystemInformation) (
    ULONG   SystemInformationClass,
    PVOID   SystemInformation,
    ULONG   SystemInformationLength,
    PULONG  ReturnLength
    );

DWORD (__stdcall *GetModuleFileNameExW) (
    HANDLE hProcess,
    HMODULE hModule,
    LPWSTR lpBaseName,
    DWORD nSize
    );

typedef LONG    NTSTATUS;
typedef LONG    KPRIORITY;

#define STATUS_INFO_LENGTH_MISMATCH      (0xC0000004L)

#define SystemProcessesAndThreadsInformation    5

typedef struct _CLIENT_ID {
    DWORD         UniqueProcess;
    DWORD         UniqueThread;
} CLIENT_ID;

typedef struct _UNICODE_STRING {
    USHORT        Length;
    USHORT        MaximumLength;
    PWSTR         Buffer;
} UNICODE_STRING;

typedef struct _VM_COUNTERS {
    SIZE_T        PeakVirtualSize;
    SIZE_T        VirtualSize;
    ULONG         PageFaultCount;
    SIZE_T        PeakWorkingSetSize;
    SIZE_T        WorkingSetSize;
    SIZE_T        QuotaPeakPagedPoolUsage;
    SIZE_T        QuotaPagedPoolUsage;
    SIZE_T        QuotaPeakNonPagedPoolUsage;
    SIZE_T        QuotaNonPagedPoolUsage;
    SIZE_T        PagefileUsage;
    SIZE_T        PeakPagefileUsage;
} VM_COUNTERS;

typedef struct _SYSTEM_THREADS {
    LARGE_INTEGER KernelTime;
    LARGE_INTEGER UserTime;
    LARGE_INTEGER CreateTime;
    ULONG         WaitTime;
    PVOID         StartAddress;
    CLIENT_ID     ClientId;
    KPRIORITY     Priority;
    KPRIORITY     BasePriority;
    ULONG         ContextSwitchCount;
    LONG          State;
    LONG          WaitReason;
} SYSTEM_THREADS, * PSYSTEM_THREADS;

typedef struct _SYSTEM_PROCESSES {
    ULONG             NextEntryDelta;
    ULONG             ThreadCount;
    ULONG             Reserved1[6];
    LARGE_INTEGER     CreateTime;
    LARGE_INTEGER     UserTime;
    LARGE_INTEGER     KernelTime;
    UNICODE_STRING    ProcessName;
    KPRIORITY         BasePriority;
    ULONG             ProcessId;
    ULONG             InheritedFromProcessId;
    ULONG             HandleCount;
    ULONG             Reserved2[2];
    VM_COUNTERS       VmCounters;
#if _WIN32_WINNT >= 0x500
    IO_COUNTERS       IoCounters;
#endif
    SYSTEM_THREADS    Threads[1];
} SYSTEM_PROCESSES, * PSYSTEM_PROCESSES;

#ifndef STATUS_BUFFER_TOO_SMALL
 #define STATUS_BUFFER_TOO_SMALL          ((NTSTATUS)0xC0000023L)
#endif // STATUS_BUFFER_TOO_SMALL

DWORD
FindProcessPid(
    WCHAR *ImagePathName
    )
{
    ULONG Status;
    ULONG cbBuffer = 0x8000;
    PVOID pBuffer = NULL;
    ULONG l=0;
    ULONG ImagePathNameLen = ImagePathName ? wcslen(ImagePathName)*sizeof(WCHAR) : 0;
    ULONG pid;

    if(!hNtdll) {
        printf("NTDLL.DLL not loaded\n");
        return 0;
    }
    if(!NtQuerySystemInformation) {
        NtQuerySystemInformation = (unsigned long (__stdcall *)(ULONG   SystemInformationClass,
                                                                PVOID   SystemInformation,
                                                                ULONG   SystemInformationLength,
                                                                PULONG  ReturnLength)) GetProcAddress( hNtdll, "NtQuerySystemInformation" );
        if(!NtQuerySystemInformation) {
            printf("NtQuerySystemInformation() not found in NTDLL.DLL\n");
            return 0;
        }
    }

    do {
        pBuffer = GlobalAlloc(GMEM_FIXED, cbBuffer);

        Status = NtQuerySystemInformation(
                    SystemProcessesAndThreadsInformation,
                    pBuffer, cbBuffer, NULL);

        if (Status == STATUS_INFO_LENGTH_MISMATCH ||
            Status == STATUS_BUFFER_TOO_SMALL) {
            GlobalFree(pBuffer);
            cbBuffer *= 2;
            if(!cbBuffer)
                return 0;
        } else
        if(Status != 0) {
            return 0;
        }
    } while (Status);

    PSYSTEM_PROCESSES pProcesses = (PSYSTEM_PROCESSES)pBuffer;

    for (;;) {

        if(exact) {
            if ((pProcesses->ProcessName.Length == ImagePathNameLen) &&
                (memicmp(pProcesses->ProcessName.Buffer, ImagePathName, ImagePathNameLen) == 0)) {
                pid = pProcesses->ProcessId;
                GlobalFree(pBuffer);
                return pid;
            }
        } else {
            for(l=0; pProcesses->ProcessName.Length-l*sizeof(WCHAR) >= ImagePathNameLen; l++) {
                if((memicmp(pProcesses->ProcessName.Buffer+l, ImagePathName, ImagePathNameLen) == 0)) {
                    pid = pProcesses->ProcessId;
                    GlobalFree(pBuffer);
                    return pid;
                }
            }
        }
        if(!pProcesses->NextEntryDelta) {
            break;
        }
        l += pProcesses->NextEntryDelta;
        if(l > cbBuffer) {
            break;
        }
        pProcesses = (PSYSTEM_PROCESSES)(((LPBYTE)pProcesses)
                        + pProcesses->NextEntryDelta);
        if(l+pProcesses->NextEntryDelta > cbBuffer) {
            break;
        }
    }

    GlobalFree(pBuffer);
    return 0;

} // end FindProcessPid()


ULONG
FindProcessModulePath(
    HANDLE hProcess,
    HMODULE hModule,
    LPWSTR lpBaseName,
    DWORD nSize
    )
{
    if(!hPsApi) {
        printf("PSAPI.DLL not loaded\n");
        return 0;
    }
    if(!GetModuleFileNameExW) {
        GetModuleFileNameExW = (DWORD (__stdcall *)(HANDLE hProcess,
                                                  HMODULE hModule,
                                                  LPWSTR lpBaseName,
                                                  DWORD nSize)) GetProcAddress( hPsApi, "GetModuleFileNameExW" );

        if(!GetModuleFileNameExW) {
            printf("GetModuleFileNameExW() not found in PSAPI.DLL\n");
            return 0;
        }
    }

    return GetModuleFileNameExW(hProcess,
                              hModule,
                              lpBaseName,
                              nSize);
} // end FindProcessModulePath()


DWORD
EnumProcessPid(
    ULONG pid
    )
{
    ULONG Status;
    ULONG cbBuffer = 0x8000;
    PVOID pBuffer = NULL;
    ULONG l=0;
    PWCHAR ModulePath = NULL;

    if(!hNtdll) {
        printf("NTDLL.DLL not loaded\n");
        return 0;
    }
    if(!NtQuerySystemInformation) {
        NtQuerySystemInformation = (unsigned long (__stdcall *)(ULONG   SystemInformationClass,
                                                                PVOID   SystemInformation,
                                                                ULONG   SystemInformationLength,
                                                                PULONG  ReturnLength)) GetProcAddress( hNtdll, "NtQuerySystemInformation" );
        if(!NtQuerySystemInformation) {
            printf("NtQuerySystemInformation() not found in NTDLL.DLL\n");
            return 0;
        }
    }

    do {
        pBuffer = GlobalAlloc(GMEM_FIXED, cbBuffer);

        Status = NtQuerySystemInformation(
                    SystemProcessesAndThreadsInformation,
                    pBuffer, cbBuffer, NULL);

        if (Status == STATUS_INFO_LENGTH_MISMATCH ||
            Status == STATUS_BUFFER_TOO_SMALL) {
            GlobalFree(pBuffer);
            cbBuffer *= 2;
            if(!cbBuffer)
                return 0;
        } else
        if(Status != 0) {
            return 0;
        }
    } while (Status);

    PSYSTEM_PROCESSES pProcesses = (PSYSTEM_PROCESSES)pBuffer;

    for (;;) {

        if(pid == -1 || pProcesses->ProcessId == pid) {

            HANDLE hProcess;

            if(!ModulePath) {
                ModulePath = (PWCHAR)GlobalAlloc(GMEM_FIXED, 0x10000);
                if(!ModulePath) {
                    GlobalFree(pBuffer);
                    return 0;
                }
            }
            ModulePath[0] = 0; // set terminator
            hProcess=OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, pProcesses->ProcessId);
            if(hProcess) {
                FindProcessModulePath(hProcess, NULL, ModulePath, 0x10000);
                CloseHandle(hProcess);
            }

            printf("%8.8x (%d)\t\t%*.*S\t%S\n",
                pProcesses->ProcessId, pProcesses->ProcessId,
                pProcesses->ProcessName.Length/sizeof(WCHAR),
                pProcesses->ProcessName.Length/sizeof(WCHAR),
                pProcesses->ProcessName.Buffer,
                ModulePath);

            if(pid != -1) {
                break;
            }
        }

        if(!pProcesses->NextEntryDelta) {
            break;
        }
        l += pProcesses->NextEntryDelta;
        if(l > cbBuffer) {
            break;
        }
        pProcesses = (PSYSTEM_PROCESSES)(((LPBYTE)pProcesses)
                        + pProcesses->NextEntryDelta);
        if(l+pProcesses->NextEntryDelta > cbBuffer) {
            break;
        }
    }

    if(ModulePath) {
        GlobalFree(ModulePath);
    }
    GlobalFree(pBuffer);
    return 0;

} // end EnumProcessPid()


int
swscanf_int(
    WCHAR* s,
    PULONG pn
    )
{
    if(!s[0]) {
        return 0;
    }
    if(s[0] == '0' && s[1] == 'x') {
        if(swscanf(s+2, L"%x", pn) != 1) {
            return 0;
        }
    } else
    if(swscanf(s, L"%d", pn) != 1) {
        return 0;
    }
    return 1;
} // end swscanf_int()


int
main (void)
{
    int argc;
    WCHAR** argv;
    WCHAR*  CmdLine;

    ULONG  pid = -1;
    ULONG  ecode = -1;
    WCHAR* WinCap = NULL;
    WCHAR* ProcName = NULL;

    int i;
    int rc = 1;

    hNtdll = GetModuleHandle("ntdll.dll");
    hPsApi = LoadLibrary("psapi.dll");

    printf("Process kill utility v0.3 (c) by Alter, 2005-2008\n"
           "Home site http://www.alter.org.ua\n");

    CmdLine = GetCommandLineW();
    argv = CommandLineToArgvW(CmdLine, &argc);

    if(argc < 2) {
        print_help();
    }

    for(i=1; i<argc; i++) {
        if(!argv[i] || argv[i][0] != '-') {
/*            if(!TargetFN) {
               TargetFN = argv[i];
            } else
            if(!LinkFN) {
               LinkFN = argv[i];
            } else {*/
                print_help();
                exit(0);
/*            }
            continue;*/
        }
        if(!wcscmp(argv[i], L"-h") ||
           !wcscmp(argv[i], L"--help")) {
            print_help();
            exit(0);
        } else
        if(!wcscmp(argv[i], L"-a") ||
           !wcscmp(argv[i], L"--all")) {
            killall = TRUE;
        } else
        if(!wcscmp(argv[i], L"-e") ||
           !wcscmp(argv[i], L"--exact")) {
            exact = TRUE;
        } else
        if(!wcscmp(argv[i], L"-i") /*||
           !wcscmp(argv[i], L"--exact")*/) {
            case_sens = FALSE;
        } else
        if(!wcscmp(argv[i], L"-l") ||
           !wcscmp(argv[i], L"--list")) {
            list_only = TRUE;
        } else
        if(!wcscmp(argv[i], L"-c") ||
           !wcscmp(argv[i], L"--caption")) {

            i++;
            if(!argv[i] || i>=argc) {
                printf("Window caption not specified\n");
                print_help();
                exit(0);
            }
            WinCap = argv[i];
        } else
        if(!wcscmp(argv[i], L"-n") ||
           !wcscmp(argv[i], L"--process-name")) {

            i++;
            if(!argv[i] || i>=argc) {
                printf("Process name (executable) not specified\n");
                print_help();
                exit(0);
            }
            exact = TRUE;
            ProcName = argv[i];
        } else
        if(!wcscmp(argv[i], L"-N") ||
           !wcscmp(argv[i], L"--process-name-substr")) {

            i++;
            if(!argv[i] || i>=argc) {
                printf("Substring of the Process name (executable) not specified\n");
                print_help();
                exit(0);
            }
            exact = FALSE;
            ProcName = argv[i];
        } else
        if(!wcscmp(argv[i], L"-p") ||
           !wcscmp(argv[i], L"--pid")) {
            i++;
            if(i>=argc) {
                printf("No Process ID\n");
                print_help();
                exit(0);
            }
            if(!swscanf_int(argv[i], &pid)) {
                printf("Invalid Process ID: %S\n", argv[i]);
                exit(0);
            }
        } else
        if(!wcscmp(argv[i], L"-x") ||
           !wcscmp(argv[i], L"--exit-code")) {
            i++;
            if(i>=argc) {
                printf("No exit code\n");
                print_help();
                exit(0);
            }
            if(!swscanf_int(argv[i], &ecode)) {
                printf("Invalid exit code: %S\n", argv[i]);
                exit(0);
            }
        } else
        {
            print_help();
            exit(0);
        }
    }

    GainAllPrivileges();

    if(list_only) {
        printf("PID\t\t\tProcess\n");
        if(WinCap) {
            //kill_by_caption(WinCap);
            printf("list mode with Window caption not implemented\n");
            exit(0);
        } else {
            EnumProcessPid(pid);
        }
    } else
    if(WinCap) {
        kill_by_caption(WinCap);
    } else
    if(pid != -1) {
        kill_by_pid(pid, ecode);
    } else
    if(ProcName) {
        while(pid = FindProcessPid(ProcName)) {
            kill_by_pid(pid, ecode);
            if(!killall)
                break;
        }
        if(!kill_count) {
            printf("No matching process (%S) found\n", ProcName);
        }
    } else
    {
        printf("Target not specified\n");
        print_help();
        exit(0);
    }

    return rc;
} // end main()
